local s = Skada
if s == nil then
	return
end

local AddonName, a = ...
local L = a.Localize
local sL = LibStub("AceLocale-3.0"):GetLocale("Skada", false)

local COMBATLOG_OBJECT_AFFILIATION_OUTSIDER = COMBATLOG_OBJECT_AFFILIATION_OUTSIDER
local COMBATLOG_OBJECT_CONTROL_PLAYER = COMBATLOG_OBJECT_CONTROL_PLAYER
local COMBATLOG_OBJECT_REACTION_MASK = COMBATLOG_OBJECT_REACTION_MASK
local GetSpellInfo = GetSpellInfo
local UnitBuff = UnitBuff
local UnitClass = UnitClass
local UnitGUID = UnitGUID
local UnitName = UnitName
local bit = bit
local math = math
local pairs = pairs
local select = select
local string = string
local wipe = wipe

local function getTable(outerTable, ...)
	for i = 1, select("#", ...) do
		local key = select(i, ...)
		outerTable = outerTable[key]
		if outerTable == nil then
			return nil
		end
	end
	return outerTable
end

local function getOrMakeTable(outerTable, ...)
	for i = 1, select("#", ...) do
		local key = select(i, ...)
		if outerTable[key] == nil then
			outerTable[key] = {}
		end
		outerTable = outerTable[key]
	end
	return outerTable
end

local function updateBars(win, columnFlags, data, getTime, isSpell, overrideMax)
	local numBars = 0
	local total = 0
	
	-- set the values for each bar
	local max = 0
	for id, info in pairs(data) do
		local amount = info.healingRequired or info.amount
		if amount and amount > 0 then
			numBars = numBars + 1
			local bar = getOrMakeTable(win.dataset, numBars)
			bar.id = info.id or id
			bar.label = info.name
			bar.value = amount
			bar.class = info.class
			if getTime then
				bar.activetime = getTime(info)
			end
			if isSpell then
				bar.icon = select(3, GetSpellInfo(bar.id))
			end
			
			max = math.max(max, amount)
			total = total + amount
		end
	end
	
	win.metadata.wipestale = true
	win.metadata.maxvalue = overrideMax or max
	while #win.dataset > numBars do
		win.dataset[#win.dataset] = nil
	end
	
	-- set the text for each bar (in a separate loop so we know %s)
	for i = 1, numBars do
		local bar = win.dataset[i]
		bar.valuetext = s:FormatValueText(
			bar.value < 1000 and bar.value or s:FormatNumber(bar.value),
			columnFlags.Healing == nil or columnFlags.Healing,
			string.format(
				"%02.1f", bar.value / math.max(1, bar.activetime or 0)),
			columnFlags.HPS,
			string.format(
				"%02.1f%%", bar.value / (overrideMax or total) * 100),
			columnFlags.Percent)
	end
end

--------------------------------------------------------- Healing Required Mode
local healingRequiredMode = s:NewModule(L["Healing Required"])
local sourceSubmenu = s:NewModule(L["Source Details"])
local spellSubmenu = s:NewModule(L["Spell Details"])
local damageTakenSubmenu = s:GetModule(sL["List of damaging spells"], true)
local buffsSubmenu = s:NewModule(L["Buff Details"])

healingRequiredMode.metadata = {
	showspots = true,
	click1 = sourceSubmenu,
	click2 = damageTakenSubmenu,
	columns = { Healing = true, HPS = true, Percent = true }
}

function healingRequiredMode:Update(win, set)
	updateBars(
		win, 
		self.metadata.columns, 
		set.players, 
		function(info)
			return s:PlayerActiveTime(set, info)
		end)
end

function healingRequiredMode:GetSetSummary(set)
	return s:FormatNumber(set.healingRequired)
end

---------------------------------------------------------------- Source Submenu
sourceSubmenu.metadata = {
	showspots = true,
	click1 = spellSubmenu,
	columns = { Healing = true, HPS = true, Percent = true }
}

function sourceSubmenu:Enter(win, id, label)
	self.targetID = id
	self.title = L["Healing done to"] .. " " .. label
end

function sourceSubmenu:Update(win, set)
	local player = s:find_player(set, self.targetID)
	updateBars(
		win, 
		self.metadata.columns, 
		player.healingRequiredSources, 
		function(info)
			return s:PlayerActiveTime(set, player)
		end)
end

----------------------------------------------------------------- Spell Submenu
spellSubmenu.metadata = {
	showspots = true,
	columns = { Healing = true, HPS = true, Percent = true }
}

function spellSubmenu:Enter(win, id, label)
	self.sourceID = id
	self.title = string.format(
		"%s %s %s %s", 
		L["Healing done to"],
		s:find_player(win:get_selected_set(), sourceSubmenu.targetID).name,
		L["by"],
		label)
end

function spellSubmenu:Update(win, set)
	local player = s:find_player(set, sourceSubmenu.targetID)
	updateBars(
		win, 
		self.metadata.columns, 
		player.healingRequiredSources[self.sourceID].spells, 
		function(info)
			return s:PlayerActiveTime(set, player)
		end,
		true)
end

----------------------------------------------------------------- Buffs Submenu
buffsSubmenu.metadata = {
--	showspots = true,
	columns = { Percent = true }
}

function buffsSubmenu:Enter(win, id, label)
	self.spellName = label
	self.title = string.format(
		"%s %s %s %s", 
		L["Tanking buffs on"],
		s:find_player(win:get_selected_set(), damageTakenSubmenu.playerid).name,
		L["during"],
		label)
end

function buffsSubmenu:Update(win, set)
	local player = s:find_player(set, damageTakenSubmenu.playerid)
	if player == nil then
		return
	end
	
	local spell = player.damagetakenspells[self.spellName]
	if spell == nil or spell.buffs == nil then
		return
	end
	
	updateBars(
		win, 
		self.metadata.columns, 
		spell.buffs, 
		nil,
		true,
		spell.count)
end

-------------------------------------------------------------------- Combat Log
local shields = {}
local multipliers = {
	[55233] = 1, -- Vampiric Blood
	[105588] = 1, -- Vampiric Brood
	[22842] = 3, -- Frenzied Regeneration
	[105737] = 3, -- Mass Regeneration
}
local buffs = {}
local interestingBuffs = {
	
	-- Death Knight
	[48707] = true, -- Anti-Magic Shell
	[81256] = true, -- Dancing Rune Weapon
	[77535] = true, -- Blood Shield
	[49222] = true, -- Bone Shield
	[55233] = true, -- Vampiric Blood
	[48792] = true, -- Icebound Fortitude
	[42650] = true, -- Army of the Dead
	
	-- Druid
--	[62600] = true, -- Savage Defense
	[22812] = true, -- Barkskin
	[22842] = true, -- Frenzied Regeneration
	[61336] = true, -- Survival Instincts
	
	-- Paladin
	[20925] = true, -- Holy Shield
	[498] = true, -- Divine Protection
	[86150] = true, -- Guardian of Ancient Kings
	[31850] = true, -- Ardent Defender
	[642] = true, -- Divine Shield
	
	-- Warrior
	[2565] = true, -- Shield Block
	[871] = true, -- Shield Wall
	[55694] = true, -- Enraged Regeneration
	[23920] = true, -- Spell Reflection
	
	-- Rogue
	[5277] = true, -- Evasion
	[1966] = true, -- Feint
	[31224] = true, -- Cloak of Shadows
	[74001] = true, -- Combat Readiness
	
	-- External
	[70940] = true, -- Divine Guardian
	[1022] = true, -- Hand of Protection
	[6940] = true, -- Hand of Sacrifice
	[33206] = true, -- Pain Suppression
	[81782] = true, -- Power Word: Barrier
	[97983] = true, -- Guardian Spirit
	
	-- Items
	[107951] = true, -- Elusive -- buff from the trinket Fire of the Deep
	[109864] = true, -- Fury of the Beast Heroic -- staff proc
	[109861] = true, -- Fury of the Beast LFR -- staff proc
	[108011] = true, -- Fury of the Beast -- staff proc
	[109776] = true, -- Master Tactician Heroic -- trinket proc
	[109774] = true, -- Master Tactician LFR -- trinket proc
	[107986] = true, -- Master Tactician -- trinket proc
	[74243] = true, -- Windwalk -- Buff from the weapon enchant
	[102667] = true, -- Veil of Lies -- Buff from the trinket
	
	-- Other
	[109634] = true -- Dream, from the extra button in Madness of Deathwing
}
local sourceInfo = {}
local targetInfo = {}
local bufferInfo = {}

local function setSourceAndTarget(
	sourceID, sourceName, sourceFlags, 
	targetID, targetName, targetFlags,
	spellName)
	
	sourceInfo.playerid = sourceID
	sourceInfo.playername = sourceName
	sourceInfo.playerflags = sourceFlags
	sourceInfo.spellname = spellName
	s:FixPets(sourceInfo)
	
	targetInfo.playerid = targetID
	targetInfo.playername = targetName
	targetInfo.playerflags = targetFlags
	s:FixPets(targetInfo)
	
	return sourceInfo.spellname
end

local function incrementInfo(table, id, name, amount)
	local info = table[id]
	local new
	if info then
		info.amount = info.amount + amount
		new = false
	else
		info = { id = id, name = name, amount = amount}
		table[id] = info
		new = true
	end
	return info, new
end

------------------------------------------------------------------- Healing
local function recordHeal(set, spellID, spellName, amount, adjustedSourceInfo)
	
	local player = s:get_player(set, targetInfo.playerid, targetInfo.playername)
	if not player then
		return
	end
	
	local required = amount
	if adjustedSourceInfo.playerid == targetInfo.playerid then
		required = 0
	end
	
	-- total
	set.healingRequired = set.healingRequired + required
	
	-- player
	player.healingRequired = player.healingRequired + required
	
	-- source
	local source, new = incrementInfo(
		player.healingRequiredSources, 
		adjustedSourceInfo.playerid, 
		adjustedSourceInfo.playername, 
		amount)
	if new then
		source.class = select(2, UnitClass(adjustedSourceInfo.playername))
		source.spells = {}
	end
	
	-- spell
	incrementInfo(source.spells, spellID, spellName, amount)
end

local function maybeRecordHeal(
	sourceID, sourceName, sourceFlags, 
	targetID, targetName, targetFlags, 
	spellID, spellName, amount)
	
	if amount <= 0 then
		return
	end
	
	spellName = setSourceAndTarget(
		sourceID, sourceName, sourceFlags, 
		targetID, targetName, targetFlags, 
		spellName)
	
	for i = 1, 99999 do
		local buffName, _, _, _, _, _, _, caster, _, _, buffID, _, _, v1, v2, v3
			= UnitBuff(targetInfo.playername, i)
		if buffName == nil then
			break
		end
		
		local multiplier = 0
		if multipliers[buffID] == 1 then
			multiplier = v1
		elseif multipliers[buffID] == 3 then
			multiplier = v3
		end
		
		if multiplier > 0 then
			local buffAmount = amount / (1 + 100 / multiplier)
			bufferInfo.playerid = UnitGUID(caster)
			bufferInfo.playername = UnitName(caster)
			recordHeal(s.current, buffID, buffName, buffAmount, bufferInfo)
			recordHeal(s.total, buffID, buffName, buffAmount, bufferInfo)
			amount = amount - buffAmount
		end
	end
	
	recordHeal(s.current, spellID, spellName, amount, sourceInfo)
	recordHeal(s.total, spellID, spellName, amount, sourceInfo)
end

local function parseHeal(_, _, 
	sourceID, sourceName, sourceFlags, 
	targetID, targetName, targetFlags, 
	spellID, spellName, _, amount, overhealing, absorbed)
	
	maybeRecordHeal(
		sourceID, sourceName, sourceFlags, 
		targetID, targetName, targetFlags, 
		spellID, spellName, amount - overhealing + absorbed)
end

--------------------------------------------------------------------- Auras
local function parseAuraApplied(_, _, 
	sourceID, sourceName, sourceFlags, 
	targetID, targetName, targetFlags, 
	spellID, spellName, _, _, amount)

	if bit.band(sourceFlags, targetFlags, COMBATLOG_OBJECT_CONTROL_PLAYER) == 0
		or bit.band(targetFlags, COMBATLOG_OBJECT_AFFILIATION_OUTSIDER) ~= 0 then
		
		return
	end
	
	if amount ~= nil 
		and bit.band(sourceFlags, COMBATLOG_OBJECT_AFFILIATION_OUTSIDER) == 0
		and bit.band(sourceFlags, targetFlags, COMBATLOG_OBJECT_REACTION_MASK) ~= 0 then
		
		getOrMakeTable(shields, sourceID, targetID)[spellID] = amount
	end
	
	if interestingBuffs[spellID] then
		getOrMakeTable(buffs, targetID)[spellID] = spellName
	end
end

local function parseAuraChanged(_, event, 
	sourceID, sourceName, sourceFlags, 
	targetID, targetName, targetFlags, 
	spellID, spellName, _, _, amount)
	
	if amount ~= nil then
		local t = getTable(shields, sourceID, targetID)
		if t ~= nil then
			if t[spellID] then
				maybeRecordHeal(
					sourceID, sourceName, sourceFlags, 
					targetID, targetName, targetFlags, 
					spellID, spellName, t[spellID] - amount)
			end
			
			if event == "SPELL_AURA_REMOVED" then
				t[spellID] = nil
			else
				t[spellID] = amount
			end
		end
	end
	
	if event == "SPELL_AURA_REMOVED" then
		local t = getTable(buffs, targetID)
		if t then
			t[spellID] = nil
		end
	end
end

-------------------------------------------------------------------- Damage
local function recordDamage(set, spellID, spellName)
	local player = s:get_player(set, targetInfo.playerid, targetInfo.playername)
	if player == nil then
		return
	end
	
	local spell = getOrMakeTable(player, "damagetakenspells", spellName)
	if not spell.id then
		spell.id = spellID
		spell.name = spellName
		spell.damage = 0
	end
	
	local num = 0
	local buffList = getOrMakeTable(spell, "buffs")
	for buffID, buffName in pairs(getOrMakeTable(buffs, targetInfo.playerid)) do
		incrementInfo(buffList, buffID, buffName, 1)
		num = num + 1
	end
	if num == 1 then
		incrementInfo(buffList, -num, num .. " " .. L["Buff"], 1)
	else
		incrementInfo(buffList, -num, num .. " " .. L["Buffs"], 1)
	end
	spell.count = (spell.count or 0) + 1
end

local function parseDamage(_, _, 
	sourceID, sourceName, sourceFlags, 
	targetID, targetName, targetFlags, 
	spellID, spellName)
	
	spellName = setSourceAndTarget(
		sourceID, sourceName, sourceFlags, 
		targetID, targetName, targetFlags, 
		spellName)
	recordDamage(s.current, spellID, spellName)
	recordDamage(s.total, spellID, spellName)
end

local function parseSwing(_, _, 
	sourceID, sourceName, sourceFlags, 
	targetID, targetName, targetFlags)
	
	parseDamage(_, _,
		sourceID, sourceName, sourceFlags, 
		targetID, targetName, targetFlags, 
		6603, sL["Attack"])
end

-------------------------------------------------------------------------- Init
function healingRequiredMode:AddSetAttributes(set)
	set.healingRequired = set.healingRequired or 0
	wipe(shields)
	wipe(buffs)
end

function healingRequiredMode:AddPlayerAttributes(player)
	player.healingRequired = player.healingRequired or 0
	getOrMakeTable(player, "healingRequiredSources")
	for _, source in pairs(player.healingRequiredSources) do
		source.healing = source.healing or 0
		for _, spell in pairs(getOrMakeTable(source, "spells")) do
			spell.healing = spell.healing or 0
		end
	end
end

function healingRequiredMode:OnEnable()
	local healFlags = { src_is_interesting = true }
	s:RegisterForCL(parseHeal, 'SPELL_HEAL', healFlags)
	s:RegisterForCL(parseHeal, 'SPELL_PERIODIC_HEAL', healFlags)
	s:RegisterForCL(parseAuraApplied, 'SPELL_AURA_APPLIED', healFlags)
	s:RegisterForCL(parseAuraChanged, 'SPELL_AURA_REFRESH', healFlags)
	s:RegisterForCL(parseAuraChanged, 'SPELL_AURA_REMOVED', healFlags)
	if damageTakenSubmenu then
		local damageFlags = { dst_is_interesting_nopets = true }
		s:RegisterForCL(parseDamage, 'DAMAGE_SHIELD', damageFlags)
		s:RegisterForCL(parseDamage, 'SPELL_DAMAGE', damageFlags)
		s:RegisterForCL(parseDamage, 'SPELL_MISSED', damageFlags)
		s:RegisterForCL(parseDamage, 'SPELL_PERIODIC_DAMAGE', damageFlags)
		s:RegisterForCL(parseDamage, 'SPELL_PERIODIC_MISSED', damageFlags)
		s:RegisterForCL(parseDamage, 'SPELL_BUILDING_DAMAGE', damageFlags)
		s:RegisterForCL(parseDamage, 'SPELL_BUILDING_MISSED', damageFlags)
		s:RegisterForCL(parseDamage, 'RANGE_DAMAGE', damageFlags)
		s:RegisterForCL(parseDamage, 'RANGE_MISSED', damageFlags)
		s:RegisterForCL(parseSwing, 'SWING_DAMAGE', damageFlags)
		s:RegisterForCL(parseSwing, 'SWING_MISSED', damageFlags)
		damageTakenSubmenu.metadata.click1 = buffsSubmenu
	end
	s:AddMode(healingRequiredMode)
end

function healingRequiredMode:OnDisable()
	s:RemoveMode(healingRequiredMode)
end
